"""Checker for pg_temp as the last entry in the search path of security definer
functions.
"""

import typing

from pglast import ast, enums as pglast_enums, visitors

from pgrubic.core import enums, linter


class SecurityDefinerFunctionTempSchemaOrder(linter.BaseChecker):
    """## **What it does**
    Checks that a security definer function has **pg_temp** as the last entry in the
    search path.

    ## **Why not?**
    Because a **SECURITY DEFINER** function is executed with the privileges of the user
    that owns it, care is needed to ensure that the function cannot be misused.
    For security, **search_path** should be set to exclude any schemas writable by
    untrusted users. This prevents malicious users from creating objects (e.g., tables,
    functions, and operators) that mask objects intended to be used by the function.
    A secure arrangement can be obtained by forcing the temporary schema to be searched
    last. To do this, write **pg_temp** as the last entry in search_path.

    ## **When should you?**
    Never.

    ## **Use instead:**
    Include **pg_temp** as the last entry in the **search_path** of SECURITY
    DEFINER functions.
    """

    is_auto_fixable = True

    def visit_CreateFunctionStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateFunctionStmt,
    ) -> None:
        """Visit a CreateFunctionStmt."""
        is_security_definer = False
        has_explicit_search_path = False
        last_entry_in_search_path = None

        for option in typing.cast(tuple[ast.DefElem], node.options):
            name = option.defname

            if name == enums.FunctionOption.SECURITY and option.arg.boolval:
                is_security_definer = True

            if (
                name == enums.FunctionOption.SET
                and isinstance(option.arg, ast.VariableSetStmt)
                and option.arg.name == "search_path"
                and option.arg.kind == pglast_enums.VariableSetKind.VAR_SET_VALUE
            ):
                has_explicit_search_path = True

                if option.arg.args:
                    last_entry_in_search_path = option.arg.args[-1].val.sval

        if (
            is_security_definer
            and has_explicit_search_path
            and last_entry_in_search_path != "pg_temp"
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Security definer function should have pg_temp as the last entry in the search path",  # noqa: E501
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Include pg_temp as the last entry in the search path",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.CreateFunctionStmt) -> None:
        """Fix violation."""
        for option in typing.cast(tuple[ast.DefElem], node.options):
            if (
                option.defname == enums.FunctionOption.SET
                and isinstance(option.arg, ast.VariableSetStmt)
                and option.arg.name == "search_path"
            ):
                non_temp_schemas = tuple(
                    schema
                    for schema in typing.cast(tuple[ast.A_Const], option.arg.args)
                    if schema.val.sval != "pg_temp"
                )

                option.arg.args = (
                    *non_temp_schemas,
                    ast.A_Const(isnull=False, val=ast.String(sval="pg_temp")),
                )



"""Checker for security definer functions without explicit search path."""

import typing

from pglast import ast, enums as pglast_enums, visitors

from pgrubic.core import enums, linter


class SecurityDefinerFunctionNoExplicitSearchPath(linter.BaseChecker):
    """## **What it does**
    Checks that a security definer function has explicit search path.

    ## **Why not?**
    Because a **SECURITY DEFINER** function is executed with the privileges of the user
    that owns it, care is needed to ensure that the function cannot be misused.
    For security, **search_path** should be set to exclude any schemas writable by
    untrusted users. This prevents malicious users from creating objects (e.g., tables,
    functions, and operators) that mask objects intended to be used by the function.

    ## **When should you?**
    Never.

    ## **Use instead:**
    Set an explicit **search_path** on SECURITY DEFINER functions.
    """

    is_auto_fixable = True

    def visit_CreateFunctionStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateFunctionStmt,
    ) -> None:
        """Visit a CreateFunctionStmt."""
        is_security_definer = False
        has_explicit_search_path = False

        for option in typing.cast(tuple[ast.DefElem], node.options):
            name = option.defname

            if name == enums.FunctionOption.SECURITY and option.arg.boolval:
                is_security_definer = True

            if (
                name == enums.FunctionOption.SET
                and isinstance(option.arg, ast.VariableSetStmt)
                and option.arg.name == "search_path"
                and option.arg.kind == pglast_enums.VariableSetKind.VAR_SET_VALUE
            ):
                has_explicit_search_path = True

        if is_security_definer and not has_explicit_search_path:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Security definer function with no explicit search path",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Set explicit search_path on security definer functions",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.CreateFunctionStmt) -> None:
        """Fix violation."""
        options_without_search_path = tuple(
            option
            for option in typing.cast(tuple[ast.DefElem], node.options)
            if not (
                option.defname == enums.FunctionOption.SET
                and isinstance(option.arg, ast.VariableSetStmt)
                and option.arg.name == "search_path"
            )
        )

        node.options = (
            *options_without_search_path,
            ast.DefElem(
                defname="set",
                arg=ast.VariableSetStmt(
                    kind=pglast_enums.VariableSetKind.VAR_SET_VALUE,
                    name="search_path",
                    args=(
                        ast.A_Const(isnull=False, val=ast.String(sval="")),
                        ast.A_Const(isnull=False, val=ast.String(sval="pg_temp")),
                    ),
                ),
            ),
        )



"""Printers for SchemaStmt."""

from pglast import ast, enums, stream, printers

from pgrubic.formatters.ddl import IF_EXISTS, IF_NOT_EXISTS


@printers.node_printer(ast.CreateSchemaStmt, override=True)
def create_schema_stmt(node: ast.CreateSchemaStmt, output: stream.RawStream) -> None:
    """Print a CreateSchemaStmt node."""
    output.writes("CREATE SCHEMA")

    if node.if_not_exists:
        output.writes(IF_NOT_EXISTS)

    if node.schemaname:
        output.print_name(node.schemaname)

    if node.authrole:
        output.newline()
        output.space(4)
        output.swrites("AUTHORIZATION")
        output.print_node(node.authrole)

    if node.schemaElts:
        output.newline()
        output.space(4)
        with output.push_indent():
            output.print_list(node.schemaElts, "", standalone_items=True)


@printers.node_printer(ast.AlterObjectSchemaStmt, override=True)
def alter_object_schema_stmt(
    node: ast.AlterObjectSchemaStmt,
    output: stream.RawStream,
) -> None:
    """Print an AlterObjectSchemaStmt node."""
    objtype = node.objectType
    output.writes("ALTER")
    output.writes(printers.ddl.OBJECT_NAMES[objtype])

    if node.missing_ok:
        output.writes(IF_EXISTS)

    if objtype in {
        enums.ObjectType.OBJECT_FOREIGN_TABLE,
        enums.ObjectType.OBJECT_MATVIEW,
        enums.ObjectType.OBJECT_SEQUENCE,
        enums.ObjectType.OBJECT_TABLE,
        enums.ObjectType.OBJECT_VIEW,
    }:
        output.print_name(node.relation)
    elif objtype in (enums.ObjectType.OBJECT_OPCLASS, enums.ObjectType.OBJECT_OPFAMILY):
        method, *name = node.object
        output.print_name(name)
        output.swrites("USING")
        output.print_symbol(method)
    else:
        output.print_name(node.object)

    output.newline()
    output.space(4)
    output.swrites("SET SCHEMA")
    output.print_name(node.newschema)



"""Formatter for function/procedure."""

from pglast import ast, enums, stream, printers

from pgrubic.core import Formatter, noqa, config

_config: config.Config = config.parse_config()


@printers.node_printer(ast.CreateFunctionStmt, override=True)
def create_function_stmt(
    node: ast.CreateFunctionStmt,
    output: stream.RawStream,
) -> None:
    """Printer for CreateFunctionStmt."""
    output.write("CREATE")
    output.space()

    if node.replace:
        output.write("OR REPLACE")
        output.space()

    if node.is_procedure:
        output.write("PROCEDURE")
    else:
        output.write("FUNCTION")

    output.space()
    output.print_name(node.funcname)
    output.space()

    if node.parameters:
        actual_parameters = node.parameters
        # When returning table, the column definitions are mixed with
        # the usual parameters, so we need to extract the column definitions
        table_definition = []

        if node.returnType and node.returnType.setof:
            # Recreate the actual parameters
            actual_parameters = []

            for param in node.parameters:
                if param.mode == enums.FunctionParameterMode.FUNC_PARAM_TABLE:
                    table_definition.append(param)
                else:
                    actual_parameters.append(param)
    else:
        actual_parameters = []
        table_definition = []

    with output.push_indent(relative=False):
        output.write("(")

        if actual_parameters:
            output.newline()
            output.space(4)
            output.print_list(actual_parameters)
            output.newline()

        output.write(")")

    if node.returnType:
        output.newline()
        output.writes("RETURNS")
        if node.returnType.setof and table_definition:
            # Do not treat them as arguments
            output.write("TABLE")
            output.space()

            with output.push_indent(relative=False):
                output.write("(")
                output.newline()
                output.space(4)
                output.print_list(table_definition)
                output.newline()
                output.write(")")
        else:
            output.print_node(node.returnType)

    if node.options:
        # Move "AS" to the end
        options = [x for x in node.options if x.defname.upper() != "AS"] + [
            x for x in node.options if x.defname.upper() == "AS"
        ]

        output.newline()

        for i, option in enumerate(options):
            output.print_node(option)
            # Add newline until the last option
            if i < len(options) - 1:
                output.newline()

    if node.sql_body:
        if node.is_procedure:
            output.newline()
            output.write("BEGIN ATOMIC")
            output.newline()
            if node.sql_body:
                output.space(4)
                with output.push_indent():
                    for i, stmt in enumerate(node.sql_body[0]):
                        output.print_node(stmt)

                        if not _config.format.new_line_before_semicolon:
                            output.write(noqa.SEMI_COLON)
                        else:  # pragma: no cover
                            output.write(noqa.NEW_LINE + noqa.SEMI_COLON)

                        # Add newline until the last statement
                        if i < len(node.sql_body[0]) - 1:
                            for _ in range(_config.format.lines_between_statements + 1):
                                output.newline()

            output.newline()
            output.write("END")
        else:
            output.newline()
            output.print_node(node.sql_body)


@printers.node_printer(
    (ast.AlterFunctionStmt, ast.CreateFunctionStmt, ast.DoStmt),
    ast.DefElem,
    override=True,
)
def create_function_option(  # noqa: PLR0911
    node: ast.CreateFunctionStmt | ast.AlterFunctionStmt | ast.DoStmt,
    output: stream.RawStream,
) -> None:
    """Printer for function options."""
    option = node.defname

    if option.upper() == "AS":
        if isinstance(node.arg, tuple) and len(node.arg) > 1:
            # This form of the AS clause is used for dynamically loadable C language
            # functions when the function name in the C language source code is not the
            # same as the name of the SQL function.
            # https://www.postgresql.org/docs/current/sql-createfunction.html#:~:text=obj_file%2C%20link_symbol
            output.write("AS")
            output.space()
            output.print_list(node.arg, standalone_items=False)
            return

        is_sql_function = False
        if isinstance(abs(node.ancestors).node, ast.CreateFunctionStmt):
            # We could use (abs(node.ancestors).node.sql_body) to check if it is a
            # sql function but sql_body is only populated for the new syntax
            # https://www.postgresql.org/docs/14/sql-createfunction.html#:~:text=a%20new%20session.-,sql_body,-The%20body%20of

            for ancestor in node.ancestors.node:
                if (
                    ancestor.defname.upper() == "LANGUAGE"
                    and ancestor.arg.sval.upper() == "SQL"
                ):
                    is_sql_function = True
                    break

            output.write("AS")
            output.space()

        function_body: str = (
            node.arg[0].sval if isinstance(node.arg, tuple) else node.arg.sval
        )

        function_body = (
            function_body + noqa.SEMI_COLON
            if not function_body.strip().endswith(noqa.SEMI_COLON)
            else function_body
        )

        output.write("$BODY$")

        if is_sql_function:
            # No error tracking is needed here because the SQL function body has already
            # been parsed successfully by the parser.
            formatted_function_body, _ = Formatter.run(
                source_code=function_body,
                source_file="function_body",
                config=_config,
            )

            # indent the non empty lines of the formatted function body by 4 spaces
            formatted_function_body = noqa.NEW_LINE.join(
                (noqa.SPACE * 4 + line if line.strip() else "")
                for line in formatted_function_body.splitlines()
            )
            output.newline()
            output.write(formatted_function_body)
            output.newline()
        else:
            output.newline()
            output.write(function_body.strip(noqa.NEW_LINE))
            output.newline()

        output.write("$BODY$")
        return

    if option.upper() == "SECURITY":
        if node.arg.boolval:
            output.swrite("SECURITY DEFINER")
        else:
            output.swrite("SECURITY INVOKER")
        return

    if option.upper() == "STRICT":
        output.swrite(
            (
                "RETURNS NULL ON NULL INPUT"
                if node.arg.boolval
                else "CALLED ON NULL INPUT"
            ),
        )
        return

    if option.upper() == "VOLATILITY":
        output.separator()
        output.write(node.arg.sval.upper())
        return

    if option.upper() == "PARALLEL":
        output.swrite("PARALLEL")
        output.space()
        output.print_symbol(node.arg.sval.upper())
        return

    if option.upper() == "LEAKPROOF":
        if not node.arg.boolval:
            output.swrite("NOT")
        output.swrite("LEAKPROOF")
        return

    if option.upper() == "SET":
        output.separator()
        output.print_node(node.arg)
        return

    if option.upper() == "WINDOW":
        output.write("WINDOW")
        return

    if option.upper() == "LANGUAGE":
        output.write("LANGUAGE")
        output.space()
        if node.arg.sval.upper() in ("SQL", "C"):
            output.print_symbol(node.arg.sval.upper())
        else:
            output.print_symbol(node.arg.sval)
        return

    output.write(node.defname.upper())
    output.space()
    output.print_symbol(node.arg)


@printers.node_printer(ast.AlterFunctionStmt, override=True)
def alter_function_stmt(node: ast.AlterFunctionStmt, output: stream.RawStream) -> None:
    """Printer for AlterFunctionStmt."""
    output.write("ALTER")
    output.space()
    if node.objtype == enums.ObjectType.OBJECT_PROCEDURE:
        output.write("PROCEDURE")
        output.space()
    else:
        output.write("FUNCTION")
        output.space()
    output.print_node(node.func)
    with output.push_indent(relative=False):
        output.newline()
        output.space(4)
        output.print_list(node.actions, noqa.SPACE, standalone_items=True)



"""Checker for new not-null column with volatile default."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class NewNotNullColumnWithVolatileDefault(linter.BaseChecker):
    """## **What it does**
    Checks new not-null column with volatile default.

    ## **Why not?**
    Adding a new column with **NOT NULL** constraint and a volatile default to an already
    populated table will have to backfill the newly added column with the default, causing
    the table to be locked, in which no other operations can be performed on the table for
    the duration of the backfill. This will cause downtime if the table is concurrently
    being accessed by other clients.

    ## **When should you?**
    If the table is empty.
    If the table is not empty but is not being concurrently accessed.

    ## **Use instead:**
    1. Create the new column, nullable.
    2. Set the default value for the newly created column.
    3. Backfill the newly created column for all existing rows.
    4. Create a check constraint: **CHECK (column IS NOT NULL) NOT VALID**.
    5. Validate the constraint.
    6. Set the column as NOT NULL.
    7. Drop the check constraint.
    """

    def visit_ColumnDef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnDef,
    ) -> None:
        """Visit ColumnDef."""
        if ancestors.find_nearest(ast.AlterTableCmd) and node.constraints:
            is_not_null = False
            has_volatile_default = False

            constraints: tuple[ast.Constraint] = node.constraints

            for constraint in constraints:
                if constraint.contype == enums.ConstrType.CONSTR_NOTNULL:
                    is_not_null = True

                if (
                    constraint.contype == enums.ConstrType.CONSTR_DEFAULT
                    and not isinstance(
                        constraint.raw_expr,
                        ast.A_Const,
                    )
                ):
                    has_volatile_default = True

            if is_not_null and has_volatile_default:
                self.violations.add(
                    linter.Violation(
                        rule_code=self.code,
                        rule_name=self.name,
                        rule_category=self.category,
                        line_number=self.line_number,
                        column_offset=self.column_offset,
                        line=self.line,
                        statement_location=self.statement_location,
                        description="New not-null column with volatile default",
                        is_auto_fixable=self.is_auto_fixable,
                        is_fix_enabled=self.is_fix_enabled,
                        help="Split the operation into multiple steps",
                    ),
                )



"""Linter."""

from __future__ import annotations

import sys
import typing
import fnmatch
import pathlib
import functools
from contextlib import contextmanager

from pglast import ast, parser, stream, visitors
from colorama import Fore, Style
from caseconverter import kebabcase

from pgrubic import ISSUES_URL, PACKAGE_NAME, DOCUMENTATION_URL
from pgrubic.core import noqa, config, errors, visitors as pgrubic_visitors, formatter

if typing.TYPE_CHECKING:
    from collections import abc  # pragma: no cover


DEFAULT_LINT_REPORT_FILE: str = f"{PACKAGE_NAME}-lint-report.md"


class FixCounter:
    """Fix counter."""

    def __init__(self) -> None:
        """Initialize variables."""
        self.counter = 0

    def add(self) -> None:
        """Increment fix counter."""
        self.counter += 1

    def reset(self) -> None:
        """Reset fix counter."""
        self.counter = 0


class Violation(typing.NamedTuple):
    """Representation of rule violation."""

    rule_code: str
    rule_name: str
    rule_category: str
    line_number: int
    column_offset: int
    line: str
    statement_location: int
    description: str
    is_auto_fixable: bool
    is_fix_enabled: bool
    help: str | None = None


class LintResult(typing.NamedTuple):
    """Lint Result."""

    source_file: str
    violations: set[Violation]
    errors: set[errors.Error]
    fixed_source_code: str | None = None


class ViolationStats(typing.NamedTuple):
    """Violation Stats."""

    total: int
    auto_fixable: int
    fix_enabled: int


class CheckerMeta(type):
    """Metaclass for Checker. This metaclass handles both runtime and subclass method
    decorations ensuring decorated methods retain their decoration even in child
    processes. It is originally created for method decorations but could be extended in
    the future.
    """

    def __new__(
        cls: type[CheckerMeta],
        name: str,
        bases: tuple[typing.Any],
        dct: dict[str, typing.Any],
    ) -> typing.Any:
        """Add method decorations."""
        for attr_name, attr_value in dct.items():
            if callable(attr_value) and attr_name.startswith("_fix"):
                dct[attr_name] = cls._apply_fix(attr_value)

            if callable(attr_value) and attr_name.startswith("visit_"):
                dct[attr_name] = cls._set_locations(attr_value)

        return super().__new__(cls, name, bases, dct)

    @staticmethod
    def _set_locations(
        func: abc.Callable[..., typing.Any],
    ) -> abc.Callable[..., typing.Any]:
        """Helper method to set locations for node."""

        @functools.wraps(func)
        def wrapper(
            checker: BaseChecker,
            ancestors: visitors.Ancestor,
            node: ast.Node,
        ) -> typing.Any:
            """Set locations for node."""
            # some nodes have location attribute which is different from node location
            # for example ast.CreateTablespaceStmt while some nodes do not carry
            # location at all.
            # If a node has no location or it is an inlined sql statement,
            # we use the length of the root statement
            if (
                hasattr(node, "location")
                and isinstance(node.location, int)
                and not checker.in_inline_sql_mode
            ):
                checker.node_location = checker.statement_location + node.location
            else:
                checker.node_location = checker.statement_location + len(
                    checker.root_statement,
                )

            # get the position of the newline just before our node location,
            line_start = (
                checker.source_code.rfind(noqa.NEW_LINE, 0, checker.node_location) + 1
            )
            # get the position of the newline just after our node location
            line_end = checker.source_code.find(noqa.NEW_LINE, checker.node_location)

            # line number is number of newlines before our node location,
            # increment by 1 to land on the actual node
            checker.line_number = (
                checker.source_code[: checker.node_location].count(noqa.NEW_LINE) + 1
            )
            # We account for a single space thus +1
            checker.column_offset = (checker.node_location - line_start) + 1

            # If a node has no location or it is an inlined sql statement,
            # we return the whole root statement instead
            if (
                hasattr(node, "location")
                and isinstance(node.location, int)
                and not checker.in_inline_sql_mode
            ):
                checker.line = checker.source_code[line_start:line_end]
            else:
                checker.line = checker.root_statement

            return func(checker, ancestors, node)

        return wrapper

    @staticmethod
    def _apply_fix(
        func: abc.Callable[..., typing.Any],
    ) -> abc.Callable[..., typing.Any]:
        """Helper method to apply fix only if it is applicable."""

        @functools.wraps(func)
        def wrapper(
            checker: BaseChecker,
            *args: typing.Any,
            **kwargs: typing.Any,
        ) -> typing.Any:
            """Apply fix only if it is applicable."""
            if not checker.config.lint.fix:
                return None

            if not checker.is_fix_applicable:
                return None

            result = func(checker, *args, **kwargs)

            checker.statement_fixes.add()
            checker.file_fixes.add()

            return result

        return wrapper


class BaseChecker(visitors.Visitor, metaclass=CheckerMeta):  # type: ignore[misc]
    """Define a lint rule, and store all the nodes that violate it."""

    # Should not be set directly
    # as they are set in __init_subclass__
    code: str
    name: str
    category: str

    # Is this rule automatically fixable?
    is_auto_fixable: bool = False

    # Attributes shared among all subclasses
    config: config.Config
    lint_ignores: list[noqa.NoQaDirective]
    source_file: str
    source_code: str

    statement_location: int
    node_location: int
    line_number: int
    column_offset: int
    statement: str
    root_statement: str
    line: str
    in_inline_sql_mode: bool = False

    # Track fixes
    statement_fixes: FixCounter
    file_fixes: FixCounter

    def __init__(self) -> None:
        """Initialize variables."""
        self.violations: set[Violation] = set()

    def __init_subclass__(cls, **kwargs: typing.Any) -> None:
        """Set code, name and category attributes for subclasses."""
        cls.code = cls.__module__.split(".")[-1]
        cls.name = kebabcase(cls.__name__)
        cls.category = cls.__module__.split(".")[-2]

    def visit(self, ancestors: visitors.Ancestor, node: ast.Node) -> None:
        """Visit the node."""

    @property
    def is_fix_enabled(self) -> bool:
        """Check if fix is enabled according to config."""
        # if a rule is not auto fixable, there is no need to check
        # if its fix is enabled, in which case, we return False
        if not self.is_auto_fixable:
            return False

        return not (
            (
                self.config.lint.fixable
                and not any(
                    fnmatch.fnmatch(self.code, pattern + "*")
                    for pattern in self.config.lint.fixable
                )
            )
            or any(
                fnmatch.fnmatch(self.code, pattern + "*")
                for pattern in self.config.lint.unfixable
            )
        )

    @property
    def is_fix_applicable(self) -> bool:
        """Check if fix can be applied."""
        if not self.is_auto_fixable:
            return False

        if not self.is_fix_enabled:
            return False

        # if the violation has been suppressed by noqa, there is no need to try to fix it
        for inline_ignore in self.lint_ignores:
            if (
                (
                    self.statement_location == inline_ignore.location
                    or self.source_file == inline_ignore.source_file
                )
                and inline_ignore.rule in (noqa.A_STAR, self.code)
                and not self.config.lint.ignore_noqa
            ):
                return False

        return True

    @classmethod
    @contextmanager
    def disable_auto_fix(
        cls,
        checkers: set[BaseChecker],
    ) -> typing.Generator[None, None, None]:
        """Disable auto fix for the given checkers."""
        original_states = [checker.is_auto_fixable for checker in checkers]

        try:
            for checker in checkers:
                checker.is_auto_fixable = False
            yield
        finally:
            for checker, original_state in zip(checkers, original_states, strict=True):
                checker.is_auto_fixable = original_state


class Linter:
    """Holds all lint rules, and runs them against a source code."""

    def __init__(
        self,
        config: config.Config,
        formatters: typing.Callable[
            [],
            set[typing.Callable[[], None]],
        ],
    ) -> None:
        """Initialize variables."""
        self.checkers: set[BaseChecker] = set()
        self.config = config
        self.formatter = formatter.Formatter(
            config=config,
            formatters=formatters,
        )

    @staticmethod
    def _skip_suppressed_violations(
        *,
        source_file: str,
        checker: BaseChecker,
        lint_ignores: list[noqa.NoQaDirective],
    ) -> None:
        """Skip suppressed violations.

        Parameters:
        ----------
        source_file: str
            Path to the source file.

        checker: BaseChecker
            Lint rule checker.

        lint_ignores: list[NoQaDirective]
            List of noqa directives.

        Returns:
        -------
        None
        """
        for inline_ignore in lint_ignores:
            suppressed_violations: set[Violation] = {
                violation
                for violation in checker.violations
                if (
                    (
                        violation.statement_location == inline_ignore.location
                        or source_file == inline_ignore.source_file
                    )
                    and (inline_ignore.rule in (noqa.A_STAR, checker.code))
                )
            }

            if suppressed_violations:
                inline_ignore.used = True

                checker.violations = {
                    violation
                    for violation in checker.violations
                    if violation not in suppressed_violations
                }

    @staticmethod
    def get_violation_stats(violations: set[Violation]) -> ViolationStats:
        """Get violation stats.

        Parameters:
        ----------
        violations: set[Violation]
            Violations to get stats from.

        Returns:
        -------
        ViolationStats
            Violation stats.
        """
        return ViolationStats(
            total=len(violations),
            auto_fixable=sum(1 for violation in violations if violation.is_auto_fixable),
            fix_enabled=sum(1 for violation in violations if violation.is_fix_enabled),
        )

    @staticmethod
    def print_violations(
        *,
        violations: set[Violation],
        source_file: str,
    ) -> None:
        """Print all violations collected by a checker.

        Parameters:
        ----------
        violations: set[Violation]
            Violations to print.
        source_file: str
            Path to the source file.

        Returns:
        -------
        None
        """
        for violation in violations:
            sys.stdout.write(
                f"{noqa.NEW_LINE}{source_file}:{violation.line_number}:{violation.column_offset}:"
                f"{noqa.SPACE}\033]8;;{DOCUMENTATION_URL}/rules/{violation.rule_category}/{violation.rule_name}{Style.RESET_ALL}{Fore.RED}{Style.BRIGHT}{violation.rule_code}\033]8;;{Style.RESET_ALL}:"
                f"{noqa.SPACE}{violation.description}{noqa.NEW_LINE}",
            )

            for idx, line in enumerate(
                violation.line.splitlines(keepends=False),
                start=violation.line_number - violation.line.count(noqa.NEW_LINE),
            ):
                sys.stdout.write(
                    f"{Fore.BLUE}{idx} | {Style.RESET_ALL}{Fore.RED}{Style.BRIGHT}{line}{Style.RESET_ALL}{noqa.NEW_LINE}",  # noqa: E501
                )
                # in order to have arrow pointing to the violation, we need to shift
                # the screen by the length of the line_number as well as 2 spaces
                # used above between the separator (|)
                (
                    sys.stdout.write(
                        noqa.SPACE
                        * (violation.column_offset + len(str(violation.line_number)) + 2)
                        + "^"
                        + noqa.NEW_LINE,
                    )
                    if idx == violation.line_number
                    else None
                )

    @staticmethod
    def generate_lint_report(
        *,
        lint_results: list[LintResult],
        report_file: str = DEFAULT_LINT_REPORT_FILE,
    ) -> None:
        """Generate a report in markdown for lint results. It overwrites the report file.

        Parameters:
        ----------
        lint_results: list[LintResult]
            List of lint results.
        report_file: str
            Path to the report file.

        Returns:
        -------
        None
        """
        total_violations = sum(
            len(lint_result.violations) for lint_result in lint_results
        )
        total_errors = sum(len(lint_result.errors) for lint_result in lint_results)

        lines: list[str] = [f"## {PACKAGE_NAME.capitalize()} Lint Report\n"]

        lines.append(f"Total violations: **{total_violations}**\n")
        lines.append(f"Total errors: **{total_errors}**\n")

        if total_violations > 0:
            lines.append(
                f"<details>\n<summary>Violations ({total_violations})</summary>\n",
            )
            lines.append("| File | Line | Col | Rule | Description | Help |")
            lines.append("|------|------|-----|------|-------------|------|")

            violation_rows = [
                f"|{noqa.SPACE}{pathlib.Path(lint_result.source_file).name}{noqa.SPACE}"
                f"|{noqa.SPACE}{violation.line_number}{noqa.SPACE}"
                f"|{noqa.SPACE}{violation.column_offset}{noqa.SPACE}"
                f"|{noqa.SPACE}[{violation.rule_code}]({DOCUMENTATION_URL}/rules/{violation.rule_category}/{violation.rule_name}){noqa.SPACE}"
                f"|{noqa.SPACE}{violation.description}{noqa.SPACE}"
                f"|{noqa.SPACE}{violation.help or '-'}{noqa.SPACE}|"
                for lint_result in lint_results
                for violation in lint_result.violations
            ]

            lines.extend(violation_rows)

            lines.append("</details>\n")

        if total_errors > 0:
            lines.append(f"<details>\n<summary>Errors ({total_errors})</summary>\n")
            lines.append("| File | Message | Hint |")
            lines.append("|------|---------|------|")

            error_rows = [
                f"|{noqa.SPACE}{pathlib.Path(lint_result.source_file).name}{noqa.SPACE}"
                f"|{noqa.SPACE}{error.message}{noqa.SPACE}"
                f"|{noqa.SPACE}{error.hint or '-'}{noqa.SPACE}|"
                for lint_result in lint_results
                for error in lint_result.errors
            ]

            lines.extend(error_rows)

            lines.append("</details>\n")

        pathlib.Path(report_file).write_text("\n".join(lines), encoding="utf-8")

    def run(self, *, source_file: str, source_code: str) -> LintResult:  # noqa: C901, PLR0912, PLR0915
        """Run rules on a source code.

        Parameters:
        ----------
        source_file: str
            Path to the source file.
        source_code: str
            Source code to lint.

        Returns:
        -------
        LintResult
            Lint result.
        """
        fixed_statements: list[str] = []

        violations: set[Violation] = set()
        _errors: set[errors.Error] = set()

        BaseChecker.source_code = source_code
        BaseChecker.source_file = source_file
        BaseChecker.config = self.config

        BaseChecker.file_fixes = FixCounter()
        BaseChecker.statement_fixes = FixCounter()

        statements = noqa.extract_statements(
            source_code=source_code,
        )

        file_lint_ignores = noqa.extract_file_lint_ignores(
            source_file=source_file,
            source_code=source_code,
        )

        lint_ignores = file_lint_ignores

        for statement in statements:
            statement_lint_ignores: list[noqa.NoQaDirective] = (
                noqa.extract_statement_lint_ignores(
                    statement=statement,
                    source_code=source_code,
                )
            )

            lint_ignores.extend(statement_lint_ignores)

            BaseChecker.lint_ignores = lint_ignores

            BaseChecker.root_statement = statement.text
            BaseChecker.statement_location = statement.start_location

            try:
                parse_tree: ast.Node = parser.parse_sql(statement.text)

                comments = noqa.extract_comments(
                    statement=statement,
                )

                inline_sql_statements = (
                    pgrubic_visitors.extract_nested_inline_sql_statements(
                        parse_tree,
                    )
                )

            except parser.ParseError as error:
                _errors.add(
                    errors.Error(
                        source_file=str(source_file),
                        source_code=source_code,
                        statement_start_location=statement.start_location + 1,
                        statement_end_location=statement.end_location,
                        statement=statement.text,
                        message=str(error),
                        hint=f"Make sure the statement is valid PostgreSQL statement. If it is, please report this issue at {ISSUES_URL}",  # noqa: E501
                    ),
                )
                fixed_statements.append(statement.text.strip(noqa.NEW_LINE))
                continue

            BaseChecker.root_statement = statement.text
            BaseChecker.statement_location = statement.start_location
            # Reset statement fixes counter per statement
            BaseChecker.statement_fixes.reset()

            # Signal that we are processing inline sql statements
            BaseChecker.in_inline_sql_mode = True

            # Temporarily disable auto fixes, we do not try to fix inline sql statements
            # inside plpgsql
            with BaseChecker.disable_auto_fix(self.checkers):
                for inline_sql_statement in inline_sql_statements:
                    BaseChecker.statement = inline_sql_statement
                    for checker in self.checkers:
                        checker.violations = set()
                        checker(parser.parse_sql(inline_sql_statement))

                        if not self.config.lint.ignore_noqa:
                            self._skip_suppressed_violations(
                                source_file=source_file,
                                checker=checker,
                                lint_ignores=lint_ignores,
                            )

                        violations.update(checker.violations)

            # We are done processing inline sql statements
            # Reset in_inline_sql_mode to false
            BaseChecker.in_inline_sql_mode = False

            BaseChecker.statement_location = statement.start_location
            BaseChecker.statement = statement.text
            for checker in self.checkers:
                checker.violations = set()

                checker(parse_tree)

                if not self.config.lint.ignore_noqa:
                    self._skip_suppressed_violations(
                        source_file=source_file,
                        checker=checker,
                        lint_ignores=lint_ignores,
                    )

                violations.update(checker.violations)

            # If the statement parse tree has been modified due to fixes,
            # we convert it to string
            if BaseChecker.statement_fixes.counter > 0:
                try:
                    fixed_statement = stream.IndentedStream(
                        comments=comments,
                        semicolon_after_last_statement=False,
                        special_functions=True,
                        separate_statements=self.config.format.lines_between_statements,
                        remove_pg_catalog_from_functions=self.config.format.remove_pg_catalog_from_functions,
                        comma_at_eoln=not (self.config.format.comma_at_beginning),
                    )(parse_tree)

                    if self.config.format.new_line_before_semicolon:
                        fixed_statement += noqa.NEW_LINE + noqa.SEMI_COLON
                    else:
                        fixed_statement += noqa.SEMI_COLON

                    fixed_statements.append(fixed_statement)
                except RecursionError as error:  # pragma: no cover
                    _errors.add(
                        errors.Error(
                            source_file=str(source_file),
                            source_code=source_code,
                            statement_start_location=statement.start_location + 1,
                            statement_end_location=statement.end_location,
                            statement=statement.text,
                            message=str(error),
                            hint="Maximum format depth exceeded, reduce deeply nested queries",  # noqa: E501
                        ),
                    )
                    fixed_statements.append(statement.text.strip(noqa.NEW_LINE))

            else:
                fixed_statements.append(statement.text.strip(noqa.NEW_LINE))

        fixed_source_code = None

        if BaseChecker.file_fixes.counter > 0:
            fixed_source_code = (
                noqa.NEW_LINE
                + (noqa.NEW_LINE * self.config.format.lines_between_statements)
            ).join(
                fixed_statements,
            ) + noqa.NEW_LINE  # final new line

        noqa.report_unused_lint_ignores(
            source_file=source_file,
            lint_ignores=lint_ignores,
        )

        return LintResult(
            source_file=source_file,
            violations=violations,
            errors=_errors,
            fixed_source_code=fixed_source_code,
        )



"""Test linter."""

import pathlib

from pgrubic import core
from pgrubic.core import noqa

SOURCE_FILE = "linter.sql"


def test_linter_violations(
    linter: core.Linter,
) -> None:
    """Test linter violations."""
    linting_result = linter.run(
        source_file=SOURCE_FILE,
        source_code="SELECT a = NULL, 10 = b;",
    )

    expected_number_of_violations: int = 2

    assert len(linting_result.violations) == expected_number_of_violations


def test_linter_suppressed_all_violations(
    linter: core.Linter,
) -> None:
    """Test linter suppressed all violations."""
    linting_result = linter.run(
        source_file=SOURCE_FILE,
        source_code="""
        -- noqa
        SELECT a = NULL, 10 = b;
        """,
    )

    assert len(linting_result.violations) == 0


def test_linter_suppressed_certain_violations(
    linter: core.Linter,
) -> None:
    """Test linter suppressed certain violations."""
    linting_result = linter.run(
        source_file=SOURCE_FILE,
        source_code="""
        -- noqa: GN024
        SELECT a = NULL, 10 = b;
        """,
    )

    assert len(linting_result.violations) == 1


def test_linter_violations_fixed_source_code(
    linter: core.Linter,
) -> None:
    """Test linter fixed sql."""
    linter.config.lint.fix = True
    linting_result = linter.run(
        source_file=SOURCE_FILE,
        source_code="""
SELECT a = NULL;
SELECT b IS NULL;
""",
    )

    assert (
        linting_result.fixed_source_code
        == f"SELECT a IS NULL;{noqa.NEW_LINE}{noqa.NEW_LINE}SELECT b IS NULL;{noqa.NEW_LINE}"  # noqa: E501
    )


def test_linter_violations_no_fixed_source_code(
    linter: core.Linter,
) -> None:
    """Test linter no fixed sql."""
    linter.config.lint.fix = False
    linting_result = linter.run(
        source_file=SOURCE_FILE,
        source_code="""SELECT a = NULL;

        SELECT b = NULL;
        """,
    )

    assert not linting_result.fixed_source_code


def test_linter_suppressed_violations_fixed_source_code(
    linter: core.Linter,
) -> None:
    """Test linter suppressed violations fixed sql."""
    linter.config.lint.fix = True
    linting_result = linter.run(
        source_file=SOURCE_FILE,
        source_code="""-- noqa: GN024
SELECT a = NULL;
""",
    )

    assert not linting_result.fixed_source_code


def test_lint_parse_error(linter: core.Linter) -> None:
    """Test parse error."""
    source_code: str = "CREATE TABLE tbl (activated);"

    linting_result = linter.run(
        source_file=SOURCE_FILE,
        source_code=source_code,
    )

    assert len(linting_result.errors) == 1


def test_new_line_before_semicolon(
    linter: core.Linter,
) -> None:
    """Test new line before semicolon."""
    linter.config.lint.fix = True
    linter.config.format.new_line_before_semicolon = True
    linting_result = linter.run(
        source_file=SOURCE_FILE,
        source_code="SELECT a = NULL;",
    )

    linter.config.format.new_line_before_semicolon = False

    assert (
        linting_result.fixed_source_code
        == f"SELECT a IS NULL{noqa.NEW_LINE};{noqa.NEW_LINE}"
    )


def test_fix_enablement(
    linter: core.Linter,
) -> None:
    """Test fix enablement."""
    linter.config.lint.fix = True
    linter.config.lint.fixable = ["GN024"]
    linting_result = linter.run(
        source_file=SOURCE_FILE,
        source_code="SELECT a = NULL;",
    )

    assert linting_result.fixed_source_code == f"SELECT a IS NULL;{noqa.NEW_LINE}"


def test_linter_generate_lint_report(
    linter: core.Linter,
    tmp_path: pathlib.Path,
) -> None:
    """Test linter lint report."""
    directory = tmp_path / "sub"
    directory.mkdir()
    report_file = directory / "report.md"

    # The ordering of violations is not guaranteed
    # So we only check for one violation here
    lint_result = linter.run(
        source_file=SOURCE_FILE,
        source_code="SELECT a = NULL; SELECT b FROM;",
    )

    linter.generate_lint_report(
        lint_results=[lint_result],
        report_file=str(report_file),
    )

    expected_lint_report = """## Pgrubic Lint Report

Total violations: **1**

Total errors: **1**

<details>
<summary>Violations (1)</summary>

| File | Line | Col | Rule | Description | Help |
|------|------|-----|------|-------------|------|
| linter.sql | 1 | 10 | [GN024](https://bolajiwahab.github.io/pgrubic/rules/general/null-comparison) | Comparison with NULL should be [IS | IS NOT] NULL | Use [IS | IS NOT] NULL |
</details>

<details>
<summary>Errors (1)</summary>

| File | Message | Hint |
|------|---------|------|
| linter.sql | syntax error at or near ";", at index 13 | Make sure the statement is valid PostgreSQL statement. If it is, please report this issue at https://github.com/bolajiwahab/pgrubic/issues |
</details>
"""  # noqa: E501

    assert report_file.read_text() == expected_lint_report



"""Formatter."""

import typing

from pglast import parser, stream

from pgrubic import ISSUES_URL
from pgrubic.core import noqa, config, errors


class FormatResult(typing.NamedTuple):
    """Format Result."""

    source_file: str
    original_source_code: str
    formatted_source_code: str
    errors: set[errors.Error]


class Formatter:
    """Format source code."""

    def __init__(
        self,
        *,
        config: config.Config,
        formatters: typing.Callable[[], set[typing.Callable[[], None]]],
    ) -> None:
        """Initialize variables."""
        self.formatters = formatters()
        self.config = config

    @staticmethod
    def run(
        *,
        source_file: str,
        source_code: str,
        config: config.Config,
    ) -> tuple[str, set[errors.Error]]:
        """Format source code.

        Parameters:
        ----------
        source_file: str
            Path to the source file.
        source_code: str
            Source code to format.

        Returns:
        -------
        tuple[str, set[errors.Error]]
            Formatted source code.
        """
        _errors: set[errors.Error] = set()

        formatted_statements: list[str] = []

        statements = noqa.extract_statements(
            source_code=source_code,
        )

        is_file_format_skip = noqa.check_file_format_skip(
            source_code=source_code,
        )

        if not is_file_format_skip:
            for statement in statements:
                if noqa.check_statement_format_skip(
                    source_code=source_code,
                    statement=statement,
                ):
                    formatted_statements.append(statement.text)
                    continue

                comments = noqa.extract_comments(
                    statement=statement,
                )

                try:
                    parser.parse_sql(statement.text)

                    formatted_statement = stream.IndentedStream(
                        comments=comments,
                        semicolon_after_last_statement=False,
                        remove_pg_catalog_from_functions=config.format.remove_pg_catalog_from_functions,
                        comma_at_eoln=not (config.format.comma_at_beginning),
                        special_functions=True,
                    )(statement.text)

                    if config.format.new_line_before_semicolon:
                        formatted_statement += noqa.NEW_LINE + noqa.SEMI_COLON
                    else:
                        formatted_statement += noqa.SEMI_COLON

                    formatted_statements.append(formatted_statement)

                except parser.ParseError as error:
                    _errors.add(
                        errors.Error(
                            source_file=str(source_file),
                            source_code=statement.text,
                            statement_start_location=statement.start_location + 1,
                            statement_end_location=statement.end_location,
                            statement=statement.text,
                            message=str(error),
                            hint=f"""Make sure the statement is valid PostgreSQL statement. If it is, please report this issue at {ISSUES_URL}{noqa.NEW_LINE}""",  # noqa: E501
                        ),
                    )
                    formatted_statements.append(statement.text.strip(noqa.NEW_LINE))

                except RecursionError as error:  # pragma: no cover
                    _errors.add(
                        errors.Error(
                            source_file=str(source_file),
                            source_code=statement.text,
                            statement_start_location=statement.start_location + 1,
                            statement_end_location=statement.end_location,
                            statement=statement.text,
                            message=str(error),
                            hint="Maximum format depth exceeded, reduce deeply nested queries",  # noqa: E501
                        ),
                    )
                    formatted_statements.append(statement.text.strip(noqa.NEW_LINE))

            return (
                noqa.NEW_LINE + (noqa.NEW_LINE * config.format.lines_between_statements)
            ).join(
                formatted_statements,
            ) + noqa.NEW_LINE, _errors

        return source_code, _errors

    def format(self, *, source_file: str, source_code: str) -> FormatResult:
        """Format source code.

        Parameters:
        ----------
        source_file: str
            Path to the source file.
        source_code: str
            Source code to format.

        Returns:
        -------
        FormatResult
            Formatted source code.
        """
        formatted_source_code, errors = self.run(
            source_file=source_file,
            source_code=source_code,
            config=self.config,
        )
        return FormatResult(
            source_file=source_file,
            original_source_code=source_code,
            formatted_source_code=formatted_source_code,
            errors=errors,
        )



"""Handling noqa comments."""

import sys
import typing
import pathlib
import dataclasses

from pglast import parser
from colorama import Fore, Style

from pgrubic import PACKAGE_NAME

A_STAR: typing.Final[str] = "*"
ASCII_SEMI_COLON: typing.Final[str] = "ASCII_59"
ASCII_OPEN_PARENTHESIS: typing.Final[str] = "ASCII_40"
ASCII_CLOSE_PARENTHESIS: typing.Final[str] = "ASCII_41"
SQL_COMMENT: typing.Final[str] = "SQL_COMMENT"
C_COMMENT: typing.Final[str] = "C_COMMENT"
BEGIN_BLOCK: typing.Final[str] = "BEGIN_P"
END_BLOCK: typing.Final[str] = "END_P"

LINT_IGNORE_DIRECTIVE: typing.Final[str] = "noqa"
FORMAT_IGNORE_DIRECTIVE: typing.Final[str] = "fmt"
SEMI_COLON: typing.Final[str] = ";"
NEW_LINE: typing.Final[str] = "\n"
SPACE: typing.Final[str] = " "


class Token(typing.NamedTuple):
    """Representation of a token."""

    start: int
    end: int
    name: str
    kind: str


class Statement(typing.NamedTuple):
    """Representation of an SQL statement."""

    start_location: int
    end_location: int
    text: str


def extract_statements(
    *,
    source_code: str,
) -> list[Statement]:
    """Extract statements from source code.

    Parameters:
    ----------
    source_code: str
        Source code to extract statements from.

    Returns:
    -------
    list[Statement]
        List of statements.
    """
    statements: list[Statement] = []

    statement_start_location = 0

    tokens: list[Token] = parser.scan(source_code)

    inside_block = False  # Tracks if we are inside BEGIN ... END block

    inside_parenthesis = False  # Tracks if we are inside parentheses (...)

    for token in tokens:
        if token.name == BEGIN_BLOCK:
            inside_block = True

        if inside_block and token.name == END_BLOCK:
            inside_block = False  # Function block ends

        if token.name == ASCII_OPEN_PARENTHESIS:
            inside_parenthesis = True

        if token.name == ASCII_CLOSE_PARENTHESIS:
            inside_parenthesis = False  # Parenthesis ends

        # Check if we have reached a semi-colon or the end of the source code
        if token.name == ASCII_SEMI_COLON or token is tokens[-1]:
            if not (inside_block or inside_parenthesis):
                # In order to include the last character, we need to increase the end
                # location by 1
                actual_end_location = token.end + 1

                statements.append(
                    Statement(
                        start_location=statement_start_location,
                        end_location=actual_end_location,
                        text=(source_code[statement_start_location:actual_end_location]),
                    ),
                )
                # Move to the next statement
                statement_start_location = actual_end_location + 1
            else:
                continue
    return statements


def _get_lint_rules_from_comment(
    comment: str,
    location: int,
    section: str,
) -> list[str]:
    """Get lint rules from comment.

    Parameters:
    ----------
    comment: str
        The comment.

    location: int
        Location of comment.

    section: str
        Section of comment.

    Returns:
    -------
    list[str]
        List of lint rules.
    """
    comment_remainder = comment.removeprefix(section)

    if not comment_remainder:
        return [A_STAR]

    rules: list[str] = [
        rule.strip()
        for rule in comment_remainder.removeprefix(":").split(",")
        if rule and comment_remainder.startswith(":")
    ]

    if not rules:
        sys.stderr.write(
            f"{Fore.YELLOW}Warning: Malformed `{LINT_IGNORE_DIRECTIVE}` directive at location {location}. Expected `{LINT_IGNORE_DIRECTIVE}: <rules>`{Style.RESET_ALL}{NEW_LINE}",  # noqa: E501
        )

    return rules


@dataclasses.dataclass(kw_only=True)
class NoQaDirective:
    """Representation of a noqa directive."""

    source_file: str | None = None
    location: int
    line_number: int
    column_offset: int
    rule: str
    used: bool = False


def extract_statement_lint_ignores(
    *,
    statement: Statement,
    source_code: str,
) -> list[NoQaDirective]:
    """Extract lint ignores from SQL statement.

    Parameters:
    ----------
    statement: Statement
        Statement to extract lint ignores from.

    source_code: str
        Source code to construct the line number and column offset of lint ignores from.

    Returns:
    -------
    list[NoQaDirective]
        List of lint ignores.
    """
    statement_lint_ignores: list[NoQaDirective] = []

    for token in typing.cast(list[Token], parser.scan(statement.text)):
        if token.name == SQL_COMMENT:
            actual_start_location = statement.start_location + token.start

            line_number = source_code[:actual_start_location].count(NEW_LINE) + 1

            # Here, we extract last comment because we can have a comment followed
            # by another comment e.g -- new table -- noqa: US005
            comment = (
                # In order to include the last character, we need to increase the end
                # location by 1
                statement.text[token.start : (token.end + 1)].split("--")[-1].strip()
            )

            if comment.startswith(LINT_IGNORE_DIRECTIVE):
                rules = _get_lint_rules_from_comment(
                    comment,
                    token.start,
                    LINT_IGNORE_DIRECTIVE,
                )

                statement_lint_ignores.extend(
                    NoQaDirective(
                        location=statement.start_location,
                        line_number=line_number,
                        # Calculate column offset as the position within the line where
                        # our directive starts
                        column_offset=(
                            actual_start_location
                            - source_code.rfind(NEW_LINE, 0, actual_start_location)
                        ),
                        rule=rule,
                    )
                    for rule in rules
                )

    return statement_lint_ignores


def extract_file_lint_ignores(
    *,
    source_file: str,
    source_code: str,
) -> list[NoQaDirective]:
    """Extract lint ignores from the start of a source file.

    Parameters:
    ----------
    source_file: str
        The source file.

    source_code: str
        Source code to extract lint ignores from.

    Returns:
    -------
    list[NoQaDirective]
        List of lint ignores.
    """
    file_ignores: list[NoQaDirective] = []

    for token in typing.cast(list[Token], parser.scan(source_code)):
        if token.start == 0 and token.name == SQL_COMMENT:
            # In order to include the last character, we need to increase the end
            # location by one
            actual_end_location = token.end + 1

            comment = (
                source_code[token.start : actual_end_location].split("--")[-1].strip()
            )

            if comment.startswith(f"{PACKAGE_NAME}: {LINT_IGNORE_DIRECTIVE}"):
                rules = _get_lint_rules_from_comment(
                    comment,
                    token.start,
                    section=f"{PACKAGE_NAME}: {LINT_IGNORE_DIRECTIVE}",
                )

                file_ignores.extend(
                    NoQaDirective(
                        source_file=source_file,
                        location=token.start,
                        line_number=1,
                        column_offset=1,
                        rule=rule,
                    )
                    for rule in rules
                )
        else:
            break

    return file_ignores


def check_file_format_skip(
    *,
    source_code: str,
) -> bool:
    """Check if formatting should be skipped for source code.

    Parameters:
    ----------
    source_code: str
        Source code to check file format skip for.

    Returns:
    -------
    bool
        True if format skip is found, False otherwise.
    """
    for token in typing.cast(list[Token], parser.scan(source_code)):
        if token.start == 0 and token.name == SQL_COMMENT:
            # In order to include the last character, we need to increase the end
            # location by one
            actual_end_location = token.end + 1

            comment = (
                source_code[token.start : actual_end_location].split("--")[-1].strip()
            )

            return bool(
                comment.startswith(f"{PACKAGE_NAME}: {FORMAT_IGNORE_DIRECTIVE}")
                and comment.removeprefix(f"{PACKAGE_NAME}: {FORMAT_IGNORE_DIRECTIVE}")
                .removeprefix(":")
                .strip()
                == "skip",
            )

    return False


def _check_statement_format_skip(
    *,
    statement: Statement,
) -> bool:
    """Check if formatting should be skipped for SQL statement.

    Parameters:
    ----------
    statement: Statement
        Statement to check if formatting should be skipped.

    Returns:
    -------
    bool
        True if format skip is found, False otherwise.
    """
    for token in typing.cast(list[Token], parser.scan(statement.text)):
        if token.name == SQL_COMMENT:
            # In order to include the last character, we need to increase the end
            # location by one
            actual_end_location = token.end + 1

            comment = (
                statement.text[token.start : actual_end_location].split("--")[-1].strip()
            )

            if (
                comment.startswith(FORMAT_IGNORE_DIRECTIVE)
                and comment.removeprefix(FORMAT_IGNORE_DIRECTIVE)
                .removeprefix(":")
                .strip()
                == "skip"
            ):
                return True

    return False


def check_statement_format_skip(
    *,
    source_code: str,
    statement: Statement,
) -> bool:
    """Check if formatting should be skipped for SQL statement base on source code or the
    statement.

    Parameters:
    ----------
    statement: Statement
        Statement to check if formatting should be skipped.

    Returns:
    -------
    bool
        True if format skip is found, False otherwise.
    """
    return check_file_format_skip(
        source_code=source_code,
    ) or _check_statement_format_skip(statement=statement)


class Comment(typing.NamedTuple):
    """Representation of an SQL comment."""

    location: int
    text: str
    at_start_of_line: bool
    continue_previous: bool


def extract_comments(*, statement: Statement) -> list[Comment]:
    """Extract comments from SQL statement.

    Parameters:
    ----------
    statement: Statement
        Statement to extract comments from.

    Returns:
    -------
    list[Comment]
        List of comments.
    """
    comments: list[Comment] = []
    # We have consciously decided to always have comments at the top of the
    # respective statement
    continue_previous = True

    for token in parser.scan(statement.text):
        if token.name in (C_COMMENT, SQL_COMMENT):
            comment = statement.text[token.start : (token.end + 1)]
            comments.append(
                Comment(
                    location=0,
                    text=comment,
                    at_start_of_line=True,
                    continue_previous=continue_previous,
                ),
            )
    return comments


def report_unused_lint_ignores(
    *,
    source_file: str,
    lint_ignores: list[NoQaDirective],
) -> None:
    """Get unused ignores.

    Parameters:
    ----------
    source_file: str
        Path to the source file.

    lint_ignores: list[NoQaDirective]
        List of noqa directives.

    Returns:
    -------
    None
    """
    for ignore in lint_ignores:
        if not ignore.used:
            sys.stdout.write(
                f"{source_file}:{ignore.line_number}:{ignore.column_offset}:"
                f" {Fore.YELLOW}Unused {LINT_IGNORE_DIRECTIVE} directive{Style.RESET_ALL}"
                f" (unused: {Fore.RED}{Style.BRIGHT}{ignore.rule}{Style.RESET_ALL}){NEW_LINE}",  # noqa: E501
            )


def add_file_level_general_lint_ignore(sources: set[pathlib.Path]) -> int:
    """Add file-level general lint ignore to the beginning of each source.

    Parameters:
    ----------
    sources: set[pathlib.Path]
        Set of source files.

    Returns:
    -------
    int
        Number of sources modified.
    """
    sources_modified = 0

    for source in sources:
        skip = False
        source_code = source.read_text()

        for token in typing.cast(list[Token], parser.scan(source_code)):
            if token.start == 0 and token.name == SQL_COMMENT:
                # In order to include the last character, we need to increase the end
                # location by one
                actual_end_location = token.end + 1

                comment = (
                    source_code[token.start : actual_end_location].split("--")[-1].strip()
                )

                if comment == f"{PACKAGE_NAME}: {LINT_IGNORE_DIRECTIVE}":
                    skip = True
                    break

        if not skip:
            source.write_text(
                f"-- {PACKAGE_NAME}: {LINT_IGNORE_DIRECTIVE}\n{source_code}",
            )
            sources_modified += 1
            continue

    return sources_modified



"""Checker for ordinal number order by."""

from pglast import ast, visitors

from pgrubic.core import linter


class OrdinalNumberOrderBy(linter.BaseChecker):
    """## **What it does**
    Checks that ORDER BY does not use numeric ordinals (e.g., ORDER BY 1).

    ## **Why not?**
    Using ordinals reduces readability, makes queries fragile when the **SELECT** list
    changes, and obscures intent.

    ## **When should you?**
    Almost never.

    ## **Use instead:**
    Explicit column names or expressions.
    """

    def visit_SelectStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.SelectStmt,
    ) -> None:
        """Visit SelectStmt."""
        if node.sortClause and any(
            isinstance(column, ast.SortBy)
            and isinstance(column.node, ast.A_Const)
            and isinstance(column.node.val, ast.Integer)
            for column in node.sortClause
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Ordinal numbers in ORDER BY",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use explicit column names or expressions instead",
                ),
            )



"""Test cli."""

import pathlib
from unittest.mock import patch

import pytest
from click import testing

from tests import TEST_FILE
from pgrubic import WORKERS_ENVIRONMENT_VARIABLE
from pgrubic.core import noqa, config, linter
from pgrubic.__main__ import cli


def test_cli_lint_file(tmp_path: pathlib.Path) -> None:
    """Test cli lint file."""
    runner = testing.CliRunner()

    sql_fail: str = "SELECT a = NULL;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql_fail)

    result = runner.invoke(cli, ["lint", str(file_fail)])

    assert result.exit_code == 1


def test_cli_lint_directory(tmp_path: pathlib.Path) -> None:
    """Test cli lint directory."""
    runner = testing.CliRunner()

    sql_fail: str = "SELECT a = NULL;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql_fail)

    result = runner.invoke(cli, ["lint", str(directory)])

    assert result.exit_code == 1


def test_cli_lint_current_directory(
    tmp_path: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Test cli lint current directory."""
    runner = testing.CliRunner()

    sql_fail: str = "SELECT a = NULL;"

    directory = tmp_path / "sub"
    directory.mkdir()
    monkeypatch.chdir(directory)

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql_fail)

    result = runner.invoke(cli, ["lint"])

    assert result.exit_code == 1


def test_cli_lint_complete_fix(tmp_path: pathlib.Path) -> None:
    """Test cli lint complete fix."""
    runner = testing.CliRunner()

    sql_fail: str = "SELECT a = NULL;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql_fail)

    result = runner.invoke(cli, ["lint", str(file_fail), "--fix"])

    assert result.exit_code == 0


def test_cli_lint_with_add_file_level_general_noqa(tmp_path: pathlib.Path) -> None:
    """Test cli lint with add_file_level_general_noqa."""
    runner = testing.CliRunner()

    sql_fail: str = "SELECT a = NULL;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql_fail)

    result = runner.invoke(cli, ["lint", str(file_fail), "--add-file-level-general-noqa"])

    assert (
        result.output
        == f"File-level general noqa directive added to 1 file(s){noqa.NEW_LINE}"
    )

    assert result.exit_code == 0


def test_cli_lint_with_generate_lint_report(tmp_path: pathlib.Path) -> None:
    """Test cli lint with generate_lint_report."""
    runner = testing.CliRunner()

    sql_fail: str = "SELECT a = NULL;"

    directory = tmp_path / "sub"
    directory.mkdir()
    report_file = linter.DEFAULT_LINT_REPORT_FILE

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql_fail)

    runner.invoke(cli, ["lint", str(file_fail), "--generate-lint-report"])

    expected_lint_report = """## Pgrubic Lint Report

Total violations: **1**

Total errors: **0**

<details>
<summary>Violations (1)</summary>

| File | Line | Col | Rule | Description | Help |
|------|------|-----|------|-------------|------|
| test.sql | 1 | 10 | [GN024](https://bolajiwahab.github.io/pgrubic/rules/general/null-comparison) | Comparison with NULL should be [IS | IS NOT] NULL | Use [IS | IS NOT] NULL |
</details>
"""  # noqa: E501

    assert pathlib.Path(report_file).read_text() == expected_lint_report

    runner.invoke(cli, ["lint", str(file_fail), "--generate-lint-report"])

    assert pathlib.Path(report_file).read_text() == expected_lint_report


def test_cli_lint_no_violations(tmp_path: pathlib.Path) -> None:
    """Test cli lint with add_file_level_general_noqa."""
    runner = testing.CliRunner()

    sql_fail: str = "SELECT a;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql_fail)

    result = runner.invoke(cli, ["lint", str(file_fail)])

    assert result.output == f"All checks passed!{noqa.NEW_LINE}"

    assert result.exit_code == 0


def test_cli_lint_verbose(tmp_path: pathlib.Path) -> None:
    """Test cli lint verbose."""
    runner = testing.CliRunner()

    sql_fail: str = "SELECT a = NULL;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql_fail)

    result = runner.invoke(cli, ["lint", str(file_fail), "--verbose"])

    assert result.exit_code == 1


def test_cli_lint_partial_fix(tmp_path: pathlib.Path) -> None:
    """Test cli lint partial fix."""
    runner = testing.CliRunner()

    sql_fail: str = "SELECT a = NULL; SELECT * FROM example;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql_fail)

    result = runner.invoke(cli, ["lint", str(file_fail), "--fix"])

    assert result.exit_code == 1


def test_cli_lint_ignore_noqa(tmp_path: pathlib.Path) -> None:
    """Test cli lint ignore noqa."""
    runner = testing.CliRunner()

    sql_fail: str = """
    -- noqa: GN024
    SELECT a = NULL;
    """

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql_fail)

    result = runner.invoke(cli, ["lint", str(file_fail), "--ignore-noqa"])

    assert result.exit_code == 1


def test_cli_lint_parse_error(tmp_path: pathlib.Path) -> None:
    """Test cli lint parse error."""
    runner = testing.CliRunner()

    sql: str = "CREATE TABLE tbl (activated);"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql)

    result = runner.invoke(cli, ["lint", str(file_fail)])

    assert result.exit_code == 1


def test_cli_lint_missing_config_error(tmp_path: pathlib.Path) -> None:
    """Test cli lint missing config error."""
    config_content = """
    [lint]
    required-columns = [
        { name = "foo", type = "text" }
    ]
    """
    runner = testing.CliRunner()

    directory = tmp_path / "sub"
    directory.mkdir()

    config_file = directory / config.CONFIG_FILE
    config_file.write_text(config_content)

    sql: str = "CREATE TABLE tbl (activated);"

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql)

    with patch("os.getcwd", return_value=str(directory)):
        result = runner.invoke(cli, ["lint", str(file_fail)])

        assert result.output == f"""Missing config key: data-type{noqa.NEW_LINE}"""
        assert result.exit_code == 1


def test_cli_lint_config_file_from_environment_variable_not_found_error(
    tmp_path: pathlib.Path,
) -> None:
    """Test cli lint config file from environment variable not found error."""
    runner = testing.CliRunner()

    directory = tmp_path / "sub"
    directory.mkdir()

    sql: str = "CREATE TABLE tbl (activated);"

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql)

    with patch.dict(
        "os.environ",
        {config.CONFIG_PATH_ENVIRONMENT_VARIABLE: "directory"},
    ):
        result = runner.invoke(cli, ["lint", str(file_fail)])

        assert (
            result.output
            == f"""Config file "pgrubic.toml" not found in the path set in the environment variable PGRUBIC_CONFIG_PATH{noqa.NEW_LINE}"""  # noqa: E501
        )

        assert result.exit_code == 1


def test_cli_lint_config_parse_error(
    tmp_path: pathlib.Path,
) -> None:
    """Test cli lint config parse error."""
    config_content = """
    [lint]
    fix =
    """
    runner = testing.CliRunner()

    directory = tmp_path / "sub"
    directory.mkdir()

    config_file = directory / config.CONFIG_FILE
    config_file.write_text(config_content)

    sql: str = "CREATE TABLE tbl (activated);"

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql)

    with patch("os.getcwd", return_value=str(directory)):
        result = runner.invoke(cli, ["lint", str(file_fail)])

        assert (
            result.output
            == f"""Error parsing configuration file "{config_file}"{noqa.NEW_LINE}"""
        )

        assert result.exit_code == 1


def test_cli_format_files(tmp_path: pathlib.Path) -> None:
    """Test cli format source files."""
    runner = testing.CliRunner()

    source_code: str = f"SELECT a = NULL;{noqa.NEW_LINE}"

    directory = tmp_path / "sub"
    directory.mkdir()

    source_1 = directory / "source_1.sql"
    source_1.write_text(source_code)

    source_2 = directory / "source_2.sql"
    source_2.write_text(source_code)

    source_3 = directory / "source_3.sql"
    source_3.write_text(source_code)

    result = runner.invoke(cli, ["format", str(source_1), str(source_2)])

    assert (
        result.output
        == f"{noqa.NEW_LINE}2 file(s) reformatted, 0 file(s) left unchanged{noqa.NEW_LINE}"  # noqa: E501
    )

    assert result.exit_code == 0

    # source_1 and source_2 are cached
    result = runner.invoke(cli, ["format", str(source_1), str(source_2)])
    assert (
        result.output
        == f"{noqa.NEW_LINE}0 file(s) reformatted, 2 file(s) left unchanged{noqa.NEW_LINE}"  # noqa: E501
    )

    assert result.exit_code == 0

    # Add a new source
    result = runner.invoke(cli, ["format", str(source_1), str(source_2), str(source_3)])
    assert (
        result.output
        == f"{noqa.NEW_LINE}1 file(s) reformatted, 2 file(s) left unchanged{noqa.NEW_LINE}"  # noqa: E501
    )

    assert result.exit_code == 0


def test_cli_format_file_verbose(tmp_path: pathlib.Path) -> None:
    """Test cli format file."""
    runner = testing.CliRunner()

    sql_pass: str = f"SELECT a = NULL;{noqa.NEW_LINE}"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_pass = directory / TEST_FILE
    file_pass.write_text(sql_pass)

    result = runner.invoke(cli, ["format", str(file_pass), "--verbose"])

    assert "Using default settings" in result.output

    assert result.exit_code == 0


def test_cli_format_directory(tmp_path: pathlib.Path) -> None:
    """Test cli format directory."""
    runner = testing.CliRunner()

    sql_pass: str = f"SELECT a = NULL;{noqa.NEW_LINE}"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_pass = directory / TEST_FILE
    file_pass.write_text(sql_pass)

    result = runner.invoke(cli, ["format", str(directory)])

    assert (
        result.output
        == f"{noqa.NEW_LINE}1 file(s) reformatted, 0 file(s) left unchanged{noqa.NEW_LINE}"  # noqa: E501
    )

    assert result.exit_code == 0


def test_cli_format_current_directory(
    tmp_path: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Test cli format current directory."""
    runner = testing.CliRunner()

    sql_pass: str = "SELECT a = NULL; SELECT * FROM example;"

    directory = tmp_path / "sub"
    directory.mkdir()
    monkeypatch.chdir(directory)

    file_pass = directory / TEST_FILE
    file_pass.write_text(sql_pass)

    result = runner.invoke(cli, ["format"])

    assert (
        result.output
        == f"{noqa.NEW_LINE}1 file(s) reformatted, 0 file(s) left unchanged{noqa.NEW_LINE}"  # noqa: E501
    )

    assert result.exit_code == 0


def test_cli_format_check(tmp_path: pathlib.Path) -> None:
    """Test cli format check."""
    runner = testing.CliRunner()

    sql_fail: str = "SELECT a = NULL; SELECT * FROM example;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql_fail)

    result = runner.invoke(cli, ["format", str(file_fail), "--check"])

    assert result.output == ""

    assert result.exit_code == 1


def test_cli_format_check_parse_error(tmp_path: pathlib.Path) -> None:
    """Test cli format check parse error."""
    runner = testing.CliRunner()

    sql_fail: str = "SELECT a =;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql_fail)

    result = runner.invoke(cli, ["format", str(file_fail), "--check"])

    assert f"1 error(s) found{noqa.NEW_LINE}" in result.output

    assert result.exit_code == 1


def test_cli_format_diff(tmp_path: pathlib.Path) -> None:
    """Test cli format check."""
    runner = testing.CliRunner()

    sql: str = "SELECT a = NULL; SELECT * FROM example;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql)

    result = runner.invoke(cli, ["format", str(file_fail), "--diff"])

    assert result.exit_code == 1


def test_cli_format_no_cache(tmp_path: pathlib.Path) -> None:
    """Test cli format with no cache."""
    runner = testing.CliRunner()

    sql: str = "SELECT a = NULL; SELECT * FROM example;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql)

    result = runner.invoke(cli, ["format", str(file_fail)])

    assert (
        result.output
        == f"{noqa.NEW_LINE}1 file(s) reformatted, 0 file(s) left unchanged{noqa.NEW_LINE}"  # noqa: E501
    )

    assert result.exit_code == 0

    # with cache read
    result = runner.invoke(cli, ["format", str(file_fail)])

    assert (
        result.output
        == f"{noqa.NEW_LINE}0 file(s) reformatted, 1 file(s) left unchanged{noqa.NEW_LINE}"  # noqa: E501
    )

    assert result.exit_code == 0

    # without cache
    result = runner.invoke(cli, ["format", str(file_fail), "--no-cache"])

    assert (
        result.output
        == f"{noqa.NEW_LINE}1 file(s) reformatted, 0 file(s) left unchanged{noqa.NEW_LINE}"  # noqa: E501
    )

    assert result.exit_code == 0


def test_cli_format_parse_error(tmp_path: pathlib.Path) -> None:
    """Test cli format parse error."""
    runner = testing.CliRunner()

    sql: str = "SELECT * FROM;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql)

    result = runner.invoke(cli, ["format", str(file_fail)])

    assert result.exit_code == 1


def test_max_workers_from_environment_variable(tmp_path: pathlib.Path) -> None:
    """Test max workers from environment variable."""
    with patch.dict(
        "os.environ",
        {WORKERS_ENVIRONMENT_VARIABLE: "1"},
    ):
        runner = testing.CliRunner()

        sql: str = "SELECT * FROM tbl;"

        directory = tmp_path / "sub"
        directory.mkdir()

        file_fail = directory / TEST_FILE
        file_fail.write_text(sql)

        result = runner.invoke(cli, ["format", str(file_fail)])

        assert result.exit_code == 0


def test_cli_format_missing_config_error(tmp_path: pathlib.Path) -> None:
    """Test cli format missing config error."""
    config_content = """
    [lint]
    required-columns = [
        { name = "foo", type = "text" }
    ]
    """
    runner = testing.CliRunner()

    directory = tmp_path / "sub"
    directory.mkdir()

    config_file = directory / config.CONFIG_FILE
    config_file.write_text(config_content)

    sql: str = "CREATE TABLE tbl (activated);"

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql)

    with patch("os.getcwd", return_value=str(directory)):
        result = runner.invoke(cli, ["format", str(file_fail)])

        assert result.output == f"Missing config key: data-type{noqa.NEW_LINE}"

        assert result.exit_code == 1


def test_cli_format_config_file_from_environment_variable_not_found_error(
    tmp_path: pathlib.Path,
) -> None:
    """Test cli format config file from environment variable not found error."""
    runner = testing.CliRunner()

    directory = tmp_path / "sub"
    directory.mkdir()

    sql: str = "CREATE TABLE tbl (activated);"

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql)

    with patch.dict(
        "os.environ",
        {config.CONFIG_PATH_ENVIRONMENT_VARIABLE: "directory"},
    ):
        result = runner.invoke(cli, ["format", str(file_fail)])

        assert (
            result.output
            == f"""Config file "pgrubic.toml" not found in the path set in the environment variable PGRUBIC_CONFIG_PATH{noqa.NEW_LINE}"""  # noqa: E501
        )

        assert result.exit_code == 1


def test_cli_format_config_parse_error(
    tmp_path: pathlib.Path,
) -> None:
    """Test cli format config parse error."""
    config_content = """
    [format]
    lines-between-statements =
    """
    runner = testing.CliRunner()

    directory = tmp_path / "sub"
    directory.mkdir()

    config_file = directory / config.CONFIG_FILE
    config_file.write_text(config_content)

    sql: str = "CREATE TABLE tbl (activated);"

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql)

    with patch("os.getcwd", return_value=str(directory)):
        result = runner.invoke(cli, ["format", str(file_fail)])

        assert (
            result.output
            == f"""Error parsing configuration file "{config_file}"{noqa.NEW_LINE}"""
        )

        assert result.exit_code == 1



"""Checker for new column with volatile default."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class NewColumnWithVolatileDefault(linter.BaseChecker):
    """## **What it does**
    Checks new column with volatile default.

    ## **Why not?**
    Adding a new column with a volatile default to an already populated table will have
    to backfill the newly added column with the default, causing the table to be locked
    and rewritten, in which no other operations can be performed on the table for
    the duration of the rewrite. This will cause downtime if the table is concurrently
    being accessed by other clients.

    ## **When should you?**
    If the table is empty.
    If the table is not empty but is not being concurrently accessed.

    ## **Use instead:**
    1. Create the new column, nullable and without the volatile default.
    2. Set the default value for the newly created column.
    3. Backfill the newly created column for all existing rows.
    """

    def visit_ColumnDef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnDef,
    ) -> None:
        """Visit ColumnDef."""
        if ancestors.find_nearest(ast.AlterTableCmd) and node.constraints:
            has_volatile_default = False

            constraints: tuple[ast.Constraint] = node.constraints

            for constraint in constraints:
                if (
                    constraint.contype == enums.ConstrType.CONSTR_DEFAULT
                    and not isinstance(
                        constraint.raw_expr,
                        ast.A_Const,
                    )
                ):
                    has_volatile_default = True
                    break

            if has_volatile_default:
                self.violations.add(
                    linter.Violation(
                        rule_code=self.code,
                        rule_name=self.name,
                        rule_category=self.category,
                        line_number=self.line_number,
                        column_offset=self.column_offset,
                        line=self.line,
                        statement_location=self.statement_location,
                        description="New column with volatile default",
                        is_auto_fixable=self.is_auto_fixable,
                        is_fix_enabled=self.is_fix_enabled,
                        help="Split the operation into multiple steps",
                    ),
                )



"""Checker for non-temp schema in search path of security definer functions."""

import typing

from pglast import ast, enums as pglast_enums, visitors

from pgrubic.core import enums, linter


class SecurityDefinerFunctionNonTempSchema(linter.BaseChecker):
    """## **What it does**
    Checks that a security definer function has non-temporary schema in the
    search path.

    ## **Why not?**
    Because a **SECURITY DEFINER** function is executed with the privileges of the user
    that owns it, care is needed to ensure that the function cannot be misused.
    For security, **search_path** should be set to exclude any schemas writable by
    untrusted users. This prevents malicious users from creating objects (e.g., tables,
    functions, and operators) that mask objects intended to be used by the function.
    A secure arrangement can be obtained by forcing the temporary schema to be searched
    last. To do this, write **pg_temp** as the last entry in search_path.

    ## **When should you?**
    Never.

    ## **Use instead:**
    Include non-temporary schema in the **search_path** of a SECURITY
    DEFINER functions.
    """

    is_auto_fixable = True

    def visit_CreateFunctionStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateFunctionStmt,
    ) -> None:
        """Visit a CreateFunctionStmt."""
        is_security_definer = False
        has_explicit_search_path = False
        has_non_temp_schema = False

        for option in typing.cast(tuple[ast.DefElem], node.options):
            name = option.defname

            if name == enums.FunctionOption.SECURITY and option.arg.boolval:
                is_security_definer = True

            if (
                name == enums.FunctionOption.SET
                and isinstance(option.arg, ast.VariableSetStmt)
                and option.arg.name == "search_path"
                and option.arg.kind == pglast_enums.VariableSetKind.VAR_SET_VALUE
            ):
                has_explicit_search_path = True

                if option.arg.args and any(
                    schema.val.sval != "pg_temp"
                    for schema in typing.cast(tuple[ast.A_Const], option.arg.args)
                ):
                    has_non_temp_schema = True

        if is_security_definer and has_explicit_search_path and not has_non_temp_schema:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Security definer function should have non-temp schema in the search path",  # noqa: E501
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Include non-temp schema in the search path",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.CreateFunctionStmt) -> None:
        """Fix violation."""
        for option in typing.cast(tuple[ast.DefElem], node.options):
            if (
                option.defname == enums.FunctionOption.SET
                and isinstance(option.arg, ast.VariableSetStmt)
                and option.arg.name == "search_path"
            ):
                option.arg.args = (
                    ast.A_Const(isnull=False, val=ast.String(sval="")),
                    ast.A_Const(isnull=False, val=ast.String(sval="pg_temp")),
                )



"""Visitors."""

import typing
from collections import deque

from pglast import ast, parser, stream, visitors, parse_plpgsql


class InlineSQLVisitor(visitors.Visitor):  # type: ignore[misc]
    """Visitor for extracting inline SQL statements from PLpgSQL and function calls."""

    def __init__(self) -> None:
        """Instantiate variables."""
        self._sql_statements: list[str] = []

    def visit_CreateFunctionStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateFunctionStmt,
    ) -> None:
        """Visit CreateFunctionStmt."""
        if node.sql_body:
            return

        _sql_statements = self._extract_sql_statements_from_plpgsql(
            parse_plpgsql(stream.RawStream()(node)),
        )
        self._sql_statements.extend(_sql_statements)

    def visit_DoStmt(self, ancestors: visitors.Ancestor, node: ast.DoStmt) -> None:
        """Visit DoStmt."""
        _sql_statements = self._extract_sql_statements_from_plpgsql(
            parse_plpgsql(stream.RawStream()(node)),
        )
        self._sql_statements.extend(_sql_statements)

    def visit_FuncCall(
        self,
        ancestors: visitors.Ancestor,
        node: ast.FuncCall,
    ) -> None:
        """Visit FuncCall."""
        if node.args:
            for arg in node.args:
                if isinstance(arg, ast.A_Const) and isinstance(arg.val, ast.String):
                    try:
                        # might not be an SQL statement.
                        statement = stream.RawStream()(arg.val.sval)
                        # if statement is PLpgSQL, it would be processed at a later pass
                        self._sql_statements.append(statement)
                    except parser.ParseError:
                        continue

    def _extract_sql_statements_from_plpgsql(
        self,
        node: dict[str, typing.Any],
    ) -> list[str]:
        """Extract SQL statements from PLpgSQL tokens using iterative breadth-first walk.

        Parameters:
        ----------
        node: dict[str, typing.Any]
            PLpgSQL tokens.

        Returns:
        -------
        list[str]
            List of SQL statements.
        """
        queue = deque([node])
        statements: list[str] = []

        while queue:
            current = queue.popleft()

            if isinstance(current, dict):
                for key, value in current.items():
                    # https://github.com/pganalyze/libpg_query/blob/17-latest/src/postgres/include/plpgsql.h#L891-L902
                    if key == "sqlstmt":
                        expr = value["PLpgSQL_expr"]
                        statements.append(expr["query"])
                    else:
                        queue.append(value)
            elif isinstance(current, list):
                queue.extend(current)

        return statements

    def get_sql_statements(self) -> list[str]:
        """Get SQL statements."""
        return self._sql_statements


def visit_inline_sql(node: ast.Node) -> list[str]:
    """Visit inline SQL."""
    inline_sql_visitor = InlineSQLVisitor()
    inline_sql_visitor(node)

    return inline_sql_visitor.get_sql_statements()


def extract_nested_inline_sql_statements(node: ast.Node) -> list[str]:
    """Extract nested inline SQL statements from PLpgSQL and function calls
    using iterative breadth-first walk.
    """
    statements: list[str] = []

    queue = deque([node])

    while queue:
        current_tree = queue.popleft()

        for statement in visit_inline_sql(current_tree):
            statements.append(statement)
            child_tree = parser.parse_sql(statement)
            queue.append(child_tree)

    return statements
